"""
G R A D I E N T - E N H A N C E D   N E U R A L   N E T W O R K S  (G E N N)

Author: Steven H. Berguin <steven.berguin@gtri.gatech.edu>

This package is distributed under New BSD license.
"""

import numpy as np

from smt.utils.neural_net.data import random_mini_batches
from smt.utils.neural_net.optimizer import Adam
from smt.utils.neural_net.activation import Tanh, Linear
from smt.utils.neural_net.bwd_prop import L_model_backward
from smt.utils.neural_net.fwd_prop import L_model_forward, L_grads_forward
from smt.utils.neural_net.loss import lse
from smt.utils.neural_net.metrics import rsquare
from smt.utils.neural_net.data import normalize_data, load_csv


# TODO: implement batch-norm (deeper networks might suffer from exploding/vanishing gradients during training)

# ------------------------------------ S U P P O R T   F U N C T I O N S -----------------------------------------------


def initialize_parameters(layer_dims=None):
    """
    Initialize neural network given topology using "He" initialization

    :param: layer_dims: neural architecture [n_0, n_1, n_2, ..., n_L] where n = number of nodes, L = number of layer
    :param: activation: the activation function to use: tanh, sigmoid, or relu (choice dependens on problem type)
    :param: regression: True = regression problem (last layer will be linear)
                        False = classification problem (last layer will be sigmoid)

    :return: parameters: dictionary containing the neural net parameters:

                    parameters["Wl"]: matrix of weights associated with layer l
                    parameters["bl"]: vector of biases associated with layer l
                    parameters["a1"]: activation function for each layer where:    -1 -- linear activation
                                                                                    0 -- sigmoid activation
                                                                                    1 -- tanh activation
                                                                                    2 -- relu activation
    """
    if not layer_dims:
        raise Exception("Neural net does have any layers")

    # Network topology
    number_layers = len(layer_dims) - 1  # input layer doesn't count

    # Parameters
    parameters = {}
    for l in range(1, number_layers + 1):
        parameters["W" + str(l)] = np.random.randn(
            layer_dims[l], layer_dims[l - 1]
        ) * np.sqrt(1.0 / layer_dims[l - 1])
        parameters["b" + str(l)] = np.zeros((layer_dims[l], 1))

    return parameters


# ------------------------------------ C L A S S -----------------------------------------------------------------------


class Model(object):
    @property
    def number_of_inputs(self):
        return self._n_x

    @property
    def number_of_outputs(self):
        return self._n_y

    @property
    def number_training_examples(self):
        return self._m

    @property
    def layer_dims(self):
        return self._layer_dims

    @property
    def activations(self):
        return self._activations

    @property
    def parameters(self):
        return self._parameters

    @property
    def training_history(self):
        return self._training_history

    @property
    def scale_factors(self):
        return self._scale_factors

    @property
    def training_data(self):
        X = self._X_norm * self._scale_factors["x"][1] + self._scale_factors["x"][0]
        Y = self._Y_norm * self._scale_factors["y"][1] + self._scale_factors["y"][0]
        J = self._J_norm * self._scale_factors["y"][1] / self._scale_factors["x"][1]
        return X, Y, J

    def __init__(self, **kwargs):
        self._parameters = dict()
        self._layer_dims = list()
        self._activations = list()
        self._training_history = dict()
        self._scale_factors = {"x": (1, 1), "y": (1, 1)}
        self._X_norm = None
        self._Y_norm = None
        self._J_norm = None
        self._n_x = None
        self._n_y = None
        self._m = None
        self._caches = list()
        self._J_caches = list()

        for name, value in kwargs.items():
            setattr(self, name, value)

    @classmethod
    def initialize(cls, n_x=None, n_y=None, deep=2, wide=12):
        layer_dims = [n_x] + [wide] * deep + [n_y]
        parameters = initialize_parameters(layer_dims)
        activations = [Tanh()] * deep + [Linear()]
        attributes = {
            "_parameters": parameters,
            "_activations": activations,
            "_layer_dims": layer_dims,
            "_n_x": n_x,
            "_n_y": n_y,
        }
        return cls(**attributes)

    def load_parameters(self, parameters):
        L = len(parameters) // 2
        deep = L - 1
        wide = parameters["W1"].shape[0]
        self._n_x = parameters["W1"].shape[1]
        self._n_y = parameters["W" + str(L)].shape[0]
        self._layer_dims = [self._n_x] + [wide] * deep + [self._n_y]
        self._activations = [Tanh()] * deep + [Linear()]
        self._parameters = parameters

    def train(
        self,
        X,
        Y,
        J=None,
        num_iterations=100,
        mini_batch_size=None,
        num_epochs=1,
        alpha=0.01,
        beta1=0.9,
        beta2=0.99,
        lambd=0.0,
        gamma=0.0,
        seed=None,
        silent=False,
    ):
        """
        Train the neural network

        :param X: matrix of shape (n_x, m) where n_x = no. of inputs, m = no. of training examples
        :param Y: matrix of shape (n_y, m) where n_y = no. of outputs
        :param J: tensor of size (n_y, n_x, m) representing the Jacobian:   dY1/dX1 = J[0][0]
                                                                            dY1/dX2 = J[0][1]
                                                                            ...
                                                                            dY2/dX1 = J[1][0]
                                                                            dY2/dX2 = J[1][1]
                                                                            ...

                   Note: to retrieve the i^th example for dY2/dX1: J[1][0][i] for all i = 1,...,m

        :param mini_batch_size: training data batches [batch_1, batch_2, ...] where batch_i = (X, Y, J)_i
        :param num_epochs: number of random passes through the entire data set (usually only used with mini-batch)
        :param alpha: learning rate
        :param beta1: parameter for ADAM optimizer
        :param beta2: parameter for ADAM optimizer
        :param lambd: regularization parameter
        :param gamma: gradient-enhancement parameter
        :param num_iterations: maximum number of optimizer iterations (per mini batch)
        :param seed: random seed in case user wants to ensure repeatability
        :param silent: don't print anything
        """
        self._load_training_data(X, Y, J)

        if not mini_batch_size:
            mini_batch_size = self.number_training_examples

        if silent:
            is_print = False
        elif mini_batch_size != 1:
            is_print = False
        else:
            is_print = True
        for e in range(num_epochs):
            self._training_history["epoch_" + str(e)] = dict()
            mini_batches = random_mini_batches(
                self._X_norm, self._Y_norm, self._J_norm, mini_batch_size, seed
            )
            for b, mini_batch in enumerate(mini_batches):
                # Get training data from this mini-batch
                X, Y, J = mini_batch

                # Optimization (learn parameters by minimizing prediction error)
                optimizer = Adam.initialize(
                    initial_guess=self._parameters,
                    cost_function=lambda p: self.cost(
                        p, self.activations, X, Y, J, lambd, gamma
                    ),
                    grad_function=lambda p: self.grad(
                        p, self.activations, X, Y, J, lambd, gamma
                    ),
                    learning_rate=alpha,
                    beta1=beta1,
                    beta2=beta2,
                )
                self._parameters = optimizer.optimize(
                    max_iter=num_iterations, is_print=is_print
                )

                # Compute average cost and print output
                avg_cost = np.mean(optimizer.cost_history).squeeze()
                self._training_history["epoch_" + str(e)][
                    "batch_" + str(b)
                ] = optimizer.cost_history

                if not silent:
                    print(
                        "epoch = {:d}, mini-batch = {:d}, avg cost = {:6.3f}".format(
                            e, b, avg_cost
                        )
                    )

    def evaluate(self, X):
        """
        Predict output(s) given inputs X.

        :param X: inputs to neural network, np array of shape (n_x, m) where n_x = no. inputs, m = no. training examples
        :return: Y: prediction, Y = np array of shape (n_y, m) where n_y = no. of outputs and m = no. of examples
        """
        assert X.shape[0] == self.number_of_inputs

        number_of_examples = X.shape[1]

        mu_x, sigma_x = self._scale_factors["x"]
        mu_y, sigma_y = self._scale_factors["y"]

        X_norm = (X - mu_x) / sigma_x

        Y_norm, _ = L_model_forward(X_norm, self.parameters, self.activations)

        Y = (Y_norm * sigma_y + mu_y).reshape(
            self.number_of_outputs, number_of_examples
        )

        return Y

    def print_parameters(self):
        """
        Print model parameters to screen for the user
        """
        for key, value in self._parameters.items():
            try:
                print("{}: {}".format(key, str(value.tolist())))
            except:
                print("{}: {}".format(key, value))

    def print_training_history(self):
        """
        Print model parameters to screen for the user
        """
        if self._training_history:
            for epoch, batches in self._training_history.items():
                for batch, history in batches.items():
                    for iteration, cost in enumerate(history):
                        print(
                            "{}, {}, iteration_{}, cost = {}".format(
                                epoch, batch, iteration, cost
                            )
                        )

    def plot_training_history(self, title="Training History", is_show_plot=True):
        """
        Plot the convergence history of the neural network learning algorithm
        """
        import matplotlib.pyplot as plt

        if self.training_history:
            if len(self.training_history.keys()) > 1:
                x_label = "epoch"
                y_label = "avg cost"
                y = []
                for epoch, batches in self.training_history.items():
                    avg_costs = []
                    for batch, values in batches.items():
                        avg_cost = np.mean(np.array(values))
                        avg_costs.append(avg_cost)
                    y.append(np.mean(np.array(avg_costs)))
                y = np.array(y)
                x = np.arange(len(y))
            elif len(self.training_history["epoch_0"]) > 1:
                x_label = "mini-batch"
                y_label = "avg cost"
                y = []
                for batch, values in self.training_history["epoch_0"].items():
                    avg_cost = np.mean(np.array(values))
                    y.append(avg_cost)
                y = np.array(y)
                x = np.arange(y.size)
            else:
                x_label = "optimizer iteration"
                y_label = "cost"
                y = np.array(self.training_history["epoch_0"]["batch_0"])
                x = np.arange(y.size)

            plt.plot(x, y)
            plt.xlabel(x_label)
            plt.ylabel(y_label)
            plt.title(title)

            if is_show_plot:
                plt.show()

    def _load_training_data(self, X, Y, J=None):
        """
        Load and normalize training data

        :param X: matrix of shape (n_x, m) where n_x = no. of inputs, m = no. of training examples
        :param Y: matrix of shape (n_y, m) where n_y = no. of outputs
        :param J: tensor of size (n_y, n_x, m) representing the Jacobian:   dY1/dX1 = J[0][0]
                                                                            dY1/dX2 = J[0][1]
                                                                            ...
                                                                            dY2/dX1 = J[1][0]
                                                                            dY2/dX2 = J[1][1]
                                                                            ...

                   Note: to retrieve the i^th example for dY2/dX1: J[1][0][i] for all i = 1,...,m
        """
        assert X.shape[1] == Y.shape[1]
        assert Y.shape[0] == Y.shape[0]
        assert X.shape[0] == self._n_x
        assert Y.shape[0] == self._n_y

        if J is not None:
            assert X.shape[1] == J.shape[2]
            assert X.shape[0] == J.shape[1]

        X_norm, Y_norm, J_norm, mu_x, sigma_x, mu_y, sigma_y = normalize_data(X, Y, J)
        self._X_norm = X_norm
        self._Y_norm = Y_norm
        self._J_norm = J_norm
        self._scale_factors["x"] = (mu_x, sigma_x)
        self._scale_factors["y"] = (mu_y, sigma_y)
        self._n_x, self._m = X.shape
        self._n_y = Y.shape[0]

    def cost(
        self,
        parameters,
        activations,
        x,
        y_true=None,
        dy_true=None,
        lambd=0.0,
        gamma=0.0,
    ):
        """
        Cost function for training

        :param x:
        :param parameters:
        :param activations:
        :param y_true:
        :param dy_true:
        :param lambd:
        :param gamma:
        :return:
        """
        y_pred, caches = L_model_forward(x, parameters, activations)
        dy_pred, dy_caches = L_grads_forward(x, parameters, activations)
        w = [value for name, value in parameters.items() if "W" in name]
        cost = lse(y_true, y_pred, lambd, w, dy_true, dy_pred, gamma)
        return cost

    def grad(
        self,
        parameters,
        activations,
        x,
        y_true=None,
        dy_true=None,
        lambd=0.0,
        gamma=0.0,
    ):
        """
        Gradient of cost function for training

        :param x:
        :param parameters:
        :param activations:
        :param y_true:
        :param dy_true:
        :param lambd:
        :param gamma:
        :return:
        """
        y_pred, caches = L_model_forward(x, parameters, activations)
        dy_pred, dy_caches = L_grads_forward(x, parameters, activations)
        grad = L_model_backward(
            y_pred, y_true, dy_pred, dy_true, caches, dy_caches, lambd, gamma
        )
        return grad

    def gradient(self, X):
        """
        Predict output(s) given inputs X.

        :param X: inputs to neural network, np array of shape (n_x, m) where n_x = no. inputs, m = no. training examples
        :return: J: prediction, J = np array of shape (n_y, n_x, m) = Jacobian
        """
        assert X.shape[0] == self.number_of_inputs

        number_of_examples = X.shape[1]

        mu_x, sigma_x = self._scale_factors["x"]
        mu_y, sigma_y = self._scale_factors["y"]

        X_norm = (X - mu_x) / sigma_x

        Y_norm, _ = L_model_forward(X_norm, self.parameters, self.activations)
        J_norm, _ = L_grads_forward(X_norm, self.parameters, self.activations)

        J = (J_norm * sigma_y / sigma_x).reshape(
            self.number_of_outputs, self.number_of_inputs, number_of_examples
        )

        return J

    def goodness_of_fit(self, X_test, Y_test, J_test=None, response=0, partial=0):
        import matplotlib.pyplot as plt
        import matplotlib.gridspec as gridspec

        assert X_test.shape[1] == Y_test.shape[1]
        assert Y_test.shape[0] == Y_test.shape[0]
        assert X_test.shape[0] == self.number_of_inputs
        assert Y_test.shape[0] == self.number_of_outputs
        if type(J_test) == np.ndarray:
            assert X_test.shape[1] == J_test.shape[2]
            assert X_test.shape[0] == J_test.shape[1]

        number_test_examples = Y_test.shape[1]

        Y_pred_test = self.evaluate(X_test)
        J_pred_test = self.gradient(X_test)

        X_train, Y_train, J_train = self.training_data

        Y_pred_train = self.evaluate(X_train)
        J_pred_train = self.gradient(X_train)

        if type(J_test) == np.ndarray:
            test = J_test[response, partial, :].reshape((1, number_test_examples))
            test_pred = J_pred_test[response, partial, :].reshape(
                (1, number_test_examples)
            )
            train = J_train[response, partial, :].reshape(
                (1, self.number_training_examples)
            )
            train_pred = J_pred_train[response, partial, :].reshape(
                (1, self.number_training_examples)
            )
            title = "Goodness of fit for dY" + str(response) + "/dX" + str(partial)
        else:
            test = Y_test[response, :].reshape((1, number_test_examples))
            test_pred = Y_pred_test[response, :].reshape((1, number_test_examples))
            train = Y_train[response, :].reshape((1, self.number_training_examples))
            train_pred = Y_pred_train[response, :].reshape(
                (1, self.number_training_examples)
            )
            title = "Goodness of fit for Y" + str(response)

        metrics = dict()
        metrics["R_squared"] = np.round(rsquare(test_pred, test), 2).squeeze()
        metrics["std_error"] = np.round(
            np.std(test_pred - test).reshape(1, 1), 2
        ).squeeze()
        metrics["avg_error"] = np.round(
            np.mean(test_pred - test).reshape(1, 1), 2
        ).squeeze()

        # Reference line
        y = np.linspace(
            min(np.min(test), np.min(train)), max(np.max(test), np.max(train)), 100
        )

        # Prepare to plot
        fig = plt.figure(figsize=(12, 6))
        fig.suptitle(title, fontsize=16)
        spec = gridspec.GridSpec(ncols=2, nrows=1, wspace=0.25)

        # Plot
        ax1 = fig.add_subplot(spec[0, 0])
        ax1.plot(y, y)
        ax1.scatter(test, test_pred, s=20, c="r")
        ax1.scatter(train, train_pred, s=100, c="k", marker="+")
        plt.legend(["perfect", "test", "train"])
        plt.xlabel("actual")
        plt.ylabel("predicted")
        plt.title("RSquare = " + str(metrics["R_squared"]))

        ax2 = fig.add_subplot(spec[0, 1])
        error = (test_pred - test).T
        weights = np.ones(error.shape) / test_pred.shape[1]
        ax2.hist(error, weights=weights, facecolor="g", alpha=0.75)
        plt.xlabel("Absolute Prediction Error")
        plt.ylabel("Probability")
        plt.title(
            "$\mu$="
            + str(metrics["avg_error"])
            + ", $\sigma=$"
            + str(metrics["std_error"])
        )
        plt.grid(True)
        plt.show()

        return metrics


def run_example(
    train_csv, test_csv, inputs, outputs, partials=None
):  # pragma: no cover
    """
    Example using 2D Rastrigin function (egg-crate-looking function)

    usage:        test_model(train_csv='train_data.csv',
                            test_csv='train_data.csv',
                            inputs=["X[0]", "X[1]"],
                            outputs=["Y[0]"],
                            partials=[["J[0][0]", "J[0][1]"]])

    :param train_csv: str, csv file name containing training data
    :param test_csv: str, csv file name containing test data
    :param inputs: list(str), csv column labels corresponding to inputs
    :param outputs: list(str), csv column labels corresponding to outputs
    :param partials: list(str), csv column labels corresponding to partials
    """

    # Sample data
    X_train, Y_train, J_train = load_csv(
        file=train_csv, inputs=inputs, outputs=outputs, partials=partials
    )
    X_test, Y_test, J_test = load_csv(
        file=test_csv, inputs=inputs, outputs=outputs, partials=partials
    )

    # Hyper-parameters
    alpha = 0.05
    beta1 = 0.90
    beta2 = 0.99
    lambd = 0.1
    gamma = 1.0
    deep = 2
    wide = 12
    mini_batch_size = None  # None = use all data as one batch
    num_iterations = 25
    num_epochs = 50

    # Training
    model = Model.initialize(
        n_x=X_train.shape[0], n_y=Y_train.shape[0], deep=deep, wide=wide
    )
    model.train(
        X=X_train,
        Y=Y_train,
        J=J_train,
        alpha=alpha,
        lambd=lambd,
        gamma=gamma,
        beta1=beta1,
        beta2=beta2,
        mini_batch_size=mini_batch_size,
        num_iterations=num_iterations,
        num_epochs=num_epochs,
        silent=False,
    )
    model.plot_training_history()
    model.goodness_of_fit(
        X_test, Y_test
    )  # model.goodness_of_fit(X_test, Y_test, J_test, partial=1)


if __name__ == "__main__":  # pragma: no cover
    run_example(
        train_csv="train_data.csv",
        test_csv="train_data.csv",
        inputs=["X[0]", "X[1]"],
        outputs=["Y[0]"],
        partials=[["J[0][0]", "J[0][1]"]],
    )



"""
G R A D I E N T - E N H A N C E D   N E U R A L   N E T W O R K S  (G E N N)

Author: Steven H. Berguin <steven.berguin@gtri.gatech.edu>

This package is distributed under New BSD license.
"""

import numpy as np

tensor = np.ndarray

EPS = np.finfo(float).eps  # small number to avoid division by zero


def compute_regularization(w, lambd=0.0):
    """
    Compute L2 norm penalty

    :param: w: the weight parameters of each layer of the neural net
    :param: lambd: float, regularization coefficient
    :return: penalty: np.ndarray of shape (1,)
    """
    lambd = max(0.0, lambd)  # ensure 0 < lambda
    penalty = 0.0
    for theta in w:
        penalty += np.squeeze(0.5 * lambd * np.sum(np.square(theta)))
    return penalty


def compute_gradient_enhancement(dy_true, dy_pred, gamma=0.0):
    """
    Compute gradient enhancement term (apply LSE to partials)

    :param: dy_pred: np ndarray of shape (n_y, n_x, m) -- predicted partials: AL' = d(AL)/dX
                                                               where n_y = # outputs, n_x = # inputs, m = # examples

    :param: dy_true: np ndarray of shape (n_y, n_x, m) -- true partials: Y' = d(Y)/dX
                                                          where n_y = # outputs, n_x = # inputs, m = # examples

    :return: loss: np.ndarray of shape (1,)
    """
    n_y, n_x, m = dy_pred.shape  # number of outputs, inputs, training examples
    loss = 0.0
    gamma = min(max(0.0, gamma), 1.0)  # ensure 0 < gamma < 1
    for k in range(0, n_y):
        for j in range(0, n_x):
            dy_j_pred = dy_pred[k, j, :].reshape(1, m)
            dy_j_true = dy_true[k, j, :].reshape(1, m)
            loss += np.squeeze(
                0.5 * gamma * np.dot((dy_j_pred - dy_j_true), (dy_j_pred - dy_j_true).T)
            )

    return loss


def lse(y_true, y_pred, lambd=0.0, w=None, dy_true=None, dy_pred=None, gamma=0.0):
    """
    Compute least squares estimator loss for regression

    :param: y_pred: np ndarray of shape (n_y, m) -- output of the forward propagation L_model_forward()
                                                    where n_y = no. outputs, m = no. examples

    :param: y_true: np ndarray of shape (n_y, m) -- true labels (classification) or values (regression)
                                               where n_y = no. outputs, m = no. examples
    :return: loss: np.ndarray of shape (1,)
    """
    n_y, m = y_true.shape  # number of outputs, training examples
    cost = 0.0
    for k in range(0, n_y):
        cost += np.squeeze(
            0.5 * np.dot((y_pred[k, :] - y_true[k, :]), (y_pred[k, :] - y_true[k, :]).T)
        )

    if w is not None:
        cost += compute_regularization(w, lambd)

    if dy_true is not None and dy_pred is not None:
        cost += compute_gradient_enhancement(dy_true, dy_pred, gamma)

    return 1.0 / m * cost


if __name__ == "__main__":  # pragma: no cover
    # Check that LSE computes correctly
    w = [np.array(1.0), np.array(2.0)]
    f = lambda x: w[0] * x + w[1] * x**2
    dfdx = lambda x: w[0] + 2 * w[1] * x
    m = 100
    lb = -5.0
    ub = 5.0
    x = np.linspace(lb, ub, m)
    y_true = f(x).reshape(1, m)
    y_pred = f(x).reshape(1, m) + 1.0
    dy_true = dfdx(x).reshape(1, 1, m)
    dy_pred = dfdx(x).reshape(1, 1, m) + 1.0
    loss = lse(
        y_true=y_true,
        y_pred=y_pred,
        dy_true=dy_true,
        dy_pred=dy_pred,
        w=w,
        lambd=1.0,
        gamma=1.0,
    )
    assert loss == 1.025



"""
Author: Remy Priem (remy.priem@onera.fr)

This package is distributed under New BSD license.
"""

from __future__ import division
import numpy as np
from scipy import linalg

from smt.utils.kriging import differences, componentwise_distance
from smt.surrogate_models.krg_based import KrgBased
from smt.utils.checks import check_support, check_nx, ensure_2d_array

"""
The Active kriging class.
"""


class MGP(KrgBased):
    name = "MGP"

    def _initialize(self):
        """
        Initialized MGP

        """
        super(MGP, self)._initialize()
        declare = self.options.declare
        declare("n_comp", 1, types=int, desc="Number of active dimensions")
        declare(
            "prior",
            {"mean": [0.0], "var": 5.0 / 4.0},
            types=dict,
            desc="Parameters for Gaussian prior of the Hyperparameters",
        )
        self.options["hyper_opt"] = "TNC"
        self.options["corr"] = "act_exp"

    def _componentwise_distance(self, dx, small=False, opt=0):
        """
        Compute the componentwise distance with respect to the correlation kernel


        Parameters
        ----------
        dx : numpy.ndarray
            Distance matrix.
        small : bool, optional
            Compute the componentwise distance in small (n_components) dimension
            or in initial dimension. The default is False.
        opt : int, optional
            useless for MGP

        Returns
        -------
        d : numpy.ndarray
            Component wise distance.

        """
        if small:
            d = componentwise_distance(dx, self.options["corr"], self.options["n_comp"])
        else:
            d = componentwise_distance(dx, self.options["corr"], self.nx)
        return d

    def _predict_values(self, x, is_acting=None):
        """
        Predict the value of the MGP for a given point

        Parameters
        ----------
        x : numpy.ndarray
            Point to compute.

        Raises
        ------
        ValueError
            The number fo dimension is not good.

        Returns
        -------
        y : numpy.ndarray
            Value of the MGP at the given point x.
        """
        n_eval, n_features = x.shape
        if n_features < self.nx:
            if n_features != self.options["n_comp"]:
                raise ValueError(
                    "dim(u) should be equal to %i" % self.options["n_comp"]
                )
            theta = np.eye(self.options["n_comp"]).reshape(
                (self.options["n_comp"] ** 2,)
            )
            # Get pairwise componentwise L1-distances to the input training set
            u = x
            x = self.get_x_from_u(u)

            u = u * self.embedding["norm"] - self.U_mean
            du = differences(u, Y=self.U_norma.copy())
            d = self._componentwise_distance(du, small=True)

            # Get an approximation of x
            x = (x - self.X_offset) / self.X_scale
            dx = differences(x, Y=self.X_norma.copy())
            d_x = self._componentwise_distance(dx)
        else:
            if n_features != self.nx:
                raise ValueError("dim(x) should be equal to %i" % self.X_scale.shape[0])
            theta = self.optimal_theta

            # Get pairwise componentwise L1-distances to the input training set
            x = (x - self.X_offset) / self.X_scale
            dx = differences(x, Y=self.X_norma.copy())
            d = self._componentwise_distance(dx)
            d_x = None

        # Compute the correlation function
        r = self._correlation_types[self.options["corr"]](theta, d, d_x=d_x).reshape(
            n_eval, self.nt
        )

        f = self._regression_types[self.options["poly"]](x)
        # Scaled predictor
        y_ = np.dot(f, self.optimal_par["beta"]) + np.dot(r, self.optimal_par["gamma"])
        # Predictor
        y = (self.y_mean + self.y_std * y_).ravel()
        return y

    def _predict_mgp_variances_base(self, x):
        """Base computation of MGP MSE used by predict_variances and predict_variances_no_uq"""
        _, n_features = x.shape

        if n_features < self.nx:
            if n_features != self.options["n_comp"]:
                raise ValueError(
                    "dim(u) should be equal to %i" % self.options["n_comp"]
                )
            u = x
            x = self.get_x_from_u(u)

            u = u * self.embedding["norm"] - self.U_mean
            x = (x - self.X_offset) / self.X_scale
        else:
            if n_features != self.nx:
                raise ValueError("dim(x) should be equal to %i" % self.X_scale.shape[0])
            u = None
            x = (x - self.X_offset) / self.X_scale

        dy = self._predict_value_derivatives_hyper(x, u)
        dMSE, MSE = self._predict_variance_derivatives_hyper(x, u)

        return MSE, dy, dMSE

    def _predict_variances(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        """
        Predict the variance of a specific point

        Parameters
        ----------
        x : numpy.ndarray
            Point to compute.

        Raises
        ------
        ValueError
            The number fo dimension is not good.

        Returns
        -------
        numpy.nd array
            MSE.
        """
        MSE, dy, dMSE = self._predict_mgp_variances_base(x)
        arg_1 = np.einsum("ij,ij->i", dy.T, linalg.solve(self.inv_sigma_R, dy).T)
        arg_2 = np.einsum("ij,ij->i", dMSE.T, linalg.solve(self.inv_sigma_R, dMSE).T)

        MGPMSE = np.zeros(x.shape[0])
        MGPMSE[MSE != 0] = (
            (4.0 / 3.0) * MSE[MSE != 0]
            + arg_1[MSE != 0]
            + (1.0 / (3.0 * MSE[MSE != 0])) * arg_2[MSE != 0]
        )
        MGPMSE[MGPMSE < 0.0] = 0.0
        return MGPMSE

    def predict_variances_no_uq(self, x):
        """Like predict_variances but without taking account hyperparameters uncertainty"""
        check_support(self, "variances")
        x = ensure_2d_array(x, "x")
        self._check_xdim(x)
        n = x.shape[0]
        x2 = np.copy(x)
        s2, _, _ = self._predict_mgp_variances_base(x)
        s2[s2 < 0.0] = 0.0
        return s2.reshape((n, self.ny))

    def _check_xdim(self, x):
        _, n_features = x.shape
        nx = self.nx
        if n_features < self.nx:
            nx = self.options["n_comp"]
        check_nx(nx, x)

    def _reduced_log_prior(self, theta, grad=False, hessian=False):
        """
        Compute the reduced log prior at given hyperparameters

        Parameters
        ----------
        theta : numpy.ndarray
            Hyperparameters.
        grad : bool, optional
            True to compuyte gradient. The default is False.
        hessian : bool, optional
            True to compute hessian. The default is False.

        Returns
        -------
        res : numpy.ndarray
            Value, gradient, hessian of the reduced log prior.

        """
        nb_theta = len(theta)

        if theta.ndim < 2:
            theta = np.atleast_2d(theta).T

        mean = np.ones((nb_theta, 1)) * self.options["prior"]["mean"]
        sig_inv = np.eye(nb_theta) / self.options["prior"]["var"]

        if grad:
            sig_inv_m = np.atleast_2d(np.sum(sig_inv, axis=0)).T
            res = -2.0 * (theta - mean) * sig_inv_m
        elif hessian:
            res = -2.0 * np.atleast_2d(np.sum(sig_inv, axis=0)).T
        else:
            res = -np.dot((theta - mean).T, sig_inv.dot(theta - mean))
        return res

    def _predict_value_derivatives_hyper(self, x, u=None):
        """
        Compute the derivatives of the mean of the GP with respect to the hyperparameters

        Parameters
        ----------
        x : numpy.ndarray
            Point to compute in initial dimension.
        u : numpy.ndarray, optional
            Point to compute in small dimension. The default is None.

        Returns
        -------
        dy : numpy.ndarray
            Derivatives of the mean of the GP with respect to the hyperparameters.

        """
        # Initialization
        n_eval, _ = x.shape

        # Get pairwise componentwise L1-distances to the input training set
        dx = differences(x, Y=self.X_norma.copy())
        d_x = self._componentwise_distance(dx)
        if u is not None:
            theta = np.eye(self.options["n_comp"]).reshape(
                (self.options["n_comp"] ** 2,)
            )

            # Get pairwise componentwise L1-distances to the input training set
            du = differences(u, Y=self.U_norma.copy())
            d = self._componentwise_distance(du, small=True)
        else:
            theta = self.optimal_theta

            # Get pairwise componentwise L1-distances to the input training set
            d = d_x
            d_x = None

        # Compute the correlation function
        r = self._correlation_types[self.options["corr"]](theta, d, d_x=d_x).reshape(
            n_eval, self.nt
        )
        # Compute the regression function
        f = self._regression_types[self.options["poly"]](x)

        dy = np.zeros((len(self.optimal_theta), n_eval))

        gamma = self.optimal_par["gamma"]
        Rinv_dR_gamma = self.optimal_par["Rinv_dR_gamma"]
        Rinv_dmu = self.optimal_par["Rinv_dmu"]

        for omega in range(len(self.optimal_theta)):
            drdomega = self._correlation_types[self.options["corr"]](
                theta, d, grad_ind=omega, d_x=d_x
            ).reshape(n_eval, self.nt)

            dbetadomega = self.optimal_par["dbeta_all"][omega]

            dy_omega = (
                f.dot(dbetadomega)
                + drdomega.dot(gamma)
                - r.dot(Rinv_dR_gamma[omega] + Rinv_dmu[omega])
            )

            dy[omega, :] = dy_omega[:, 0]

        return dy

    def _predict_variance_derivatives_hyper(self, x, u=None):
        """
        Compute the derivatives of the variance of the GP with respect to the hyperparameters

        Parameters
        ----------
        x : numpy.ndarray
            Point to compute in initial dimension.
        u : numpy.ndarray, optional
            Point to compute in small dimension. The default is None.

        Returns
        -------
        dMSE : numpy.ndarrray
            derivatives of the variance of the GP with respect to the hyperparameters.
        MSE : TYPE
            Variance of the GP.

        """
        # Initialization
        n_eval, n_features_x = x.shape

        # Get pairwise componentwise L1-distances to the input training set
        dx = differences(x, Y=self.X_norma.copy())
        d_x = self._componentwise_distance(dx)
        if u is not None:
            theta = np.eye(self.options["n_comp"]).reshape(
                (self.options["n_comp"] ** 2,)
            )
            # Get pairwise componentwise L1-distances to the input training set
            du = differences(u, Y=self.U_norma.copy())
            d = self._componentwise_distance(du, small=True)
        else:
            theta = self.optimal_theta
            # Get pairwise componentwise L1-distances to the input training set
            d = d_x
            d_x = None

        # Compute the correlation function
        r = (
            self._correlation_types[self.options["corr"]](theta, d, d_x=d_x)
            .reshape(n_eval, self.nt)
            .T
        )
        f = self._regression_types[self.options["poly"]](x).T

        C = self.optimal_par["C"]
        G = self.optimal_par["G"]
        Ft = self.optimal_par["Ft"]
        sigma2 = self.optimal_par["sigma2"]

        rt = linalg.solve_triangular(C, r, lower=True)

        F_Rinv_r = np.dot(Ft.T, rt)

        u_ = linalg.solve_triangular(G.T, f - F_Rinv_r)

        MSE = self.optimal_par["sigma2"] * (
            1.0 - (rt**2.0).sum(axis=0) + (u_**2.0).sum(axis=0)
        )
        # Mean Squared Error might be slightly negative depending on
        # machine precision: force to zero!
        MSE[MSE < 0.0] = 0.0

        Ginv_u = linalg.solve_triangular(G, u_, lower=False)
        Rinv_F = linalg.solve_triangular(C.T, Ft, lower=False)
        Rinv_r = linalg.solve_triangular(C.T, rt, lower=False)
        Rinv_F_Ginv_u = Rinv_F.dot(Ginv_u)

        dMSE = np.zeros((len(self.optimal_theta), n_eval))

        dr_all = self.optimal_par["dr"]
        dsigma = self.optimal_par["dsigma"]

        for omega in range(len(self.optimal_theta)):
            drdomega = (
                self._correlation_types[self.options["corr"]](
                    theta, d, grad_ind=omega, d_x=d_x
                )
                .reshape(n_eval, self.nt)
                .T
            )

            dRdomega = np.zeros((self.nt, self.nt))
            dRdomega[self.ij[:, 0], self.ij[:, 1]] = dr_all[omega][:, 0]
            dRdomega[self.ij[:, 1], self.ij[:, 0]] = dr_all[omega][:, 0]

            # Compute du2dtheta

            dRdomega_Rinv_F_Ginv_u = dRdomega.dot(Rinv_F_Ginv_u)
            r_Rinv_dRdomega_Rinv_F_Ginv_u = np.einsum(
                "ij,ij->i", Rinv_r.T, dRdomega_Rinv_F_Ginv_u.T
            )
            drdomega_Rinv_F_Ginv_u = np.einsum("ij,ij->i", drdomega.T, Rinv_F_Ginv_u.T)
            u_Ginv_F_Rinv_dRdomega_Rinv_F_Ginv_u = np.einsum(
                "ij,ij->i", Rinv_F_Ginv_u.T, dRdomega_Rinv_F_Ginv_u.T
            )

            du2domega = (
                2.0 * r_Rinv_dRdomega_Rinv_F_Ginv_u
                - 2.0 * drdomega_Rinv_F_Ginv_u
                + u_Ginv_F_Rinv_dRdomega_Rinv_F_Ginv_u
            )
            du2domega = np.atleast_2d(du2domega)

            # Compute drt2dtheta
            drdomega_Rinv_r = np.einsum("ij,ij->i", drdomega.T, Rinv_r.T)
            r_Rinv_dRdomega_Rinv_r = np.einsum(
                "ij,ij->i", Rinv_r.T, dRdomega.dot(Rinv_r).T
            )

            drt2domega = 2.0 * drdomega_Rinv_r - r_Rinv_dRdomega_Rinv_r
            drt2domega = np.atleast_2d(drt2domega)

            dMSE[omega] = dsigma[omega] * MSE / sigma2 + sigma2 * (
                -drt2domega + du2domega
            )

        return dMSE, MSE

    def get_x_from_u(self, u):
        """
        Compute the point in initial dimension from a point in low dimension

        Parameters
        ----------
        u : numpy.ndarray
            Point in low dimension.

        Returns
        -------
        res : numpy.ndarray
            point in initial dimension.

        """
        u = np.atleast_2d(u)

        self.embedding["Q_C"], self.embedding["R_C"]

        x_temp = np.dot(
            self.embedding["Q_C"],
            linalg.solve_triangular(self.embedding["R_C"].T, u.T, lower=True),
        ).T

        res = np.atleast_2d(x_temp)
        return res

    def get_u_from_x(self, x):
        """
        Compute the point in low dimension from a point in initial dimension

        Parameters
        ----------
        x : numpy.ndarray
            Point in initial dimension.

        Returns
        -------
        u : numpy.ndarray
             Point in low dimension.

        """
        u = x.dot(self.embedding["C"])
        return u

    def _specific_train(self):
        """
        Compute the specific training values necessary for MGP (Hessian)
        """
        # Compute covariance matrix of hyperparameters
        var_R = np.zeros((len(self.optimal_theta), len(self.optimal_theta)))
        r, r_ij, par = self._reduced_likelihood_hessian(self.optimal_theta)
        var_R[r_ij[:, 0], r_ij[:, 1]] = r[:, 0]
        var_R[r_ij[:, 1], r_ij[:, 0]] = r[:, 0]

        self.inv_sigma_R = -var_R

        # Compute normalise embedding
        self.optimal_par = par

        A = np.reshape(self.optimal_theta, (self.options["n_comp"], self.nx)).T
        B = (A.T / self.X_scale).T
        norm_B = np.linalg.norm(B)
        C = B / norm_B

        self.embedding = {}
        self.embedding["A"] = A
        self.embedding["C"] = C
        self.embedding["norm"] = norm_B
        self.embedding["Q_C"], self.embedding["R_C"] = linalg.qr(C, mode="economic")

        # Compute normalisation in embeding base
        self.U_norma = self.X_norma.dot(A)
        self.U_mean = self.X_offset.dot(C) * norm_B

        # Compute best number of Components for Active Kriging
        svd = linalg.svd(A)
        svd_cumsum = np.cumsum(svd[1])
        svd_sum = np.sum(svd[1])
        self.best_ncomp = min(np.argwhere(svd_cumsum > 0.99 * svd_sum)) + 1

    def _check_param(self):
        """
        Overrides KrgBased implementation
        This function checks some parameters of the model.
        """

        d = self.options["n_comp"] * self.nx

        if self.options["corr"] != "act_exp":
            raise ValueError("MGP must be used with act_exp correlation function")
        if self.options["hyper_opt"] != "TNC":
            raise ValueError("MGP must be used with TNC hyperparameters optimizer")

        if len(self.options["theta0"]) != d:
            if len(self.options["theta0"]) == 1:
                self.options["theta0"] *= np.ones(d)
            else:
                raise ValueError(
                    "the number of dim %s should be equal to the length of theta0 %s."
                    % (d, len(self.options["theta0"]))
                )



"""
Author: P.Saves 
This package is distributed under New BSD license.

Cantilever beam problem from:
P. Saves, Y. Diouane, N. Bartoli, T. Lefebvre, and J. Morlier. A mixed-categorical correlation kernel for gaussian process, 2022
"""
import numpy as np

from smt.problems.problem import Problem
from smt.utils.design_space import DesignSpace, FloatVariable, CategoricalVariable


class MixedCantileverBeam(Problem):
    def _initialize(self):
        self.options.declare("name", "MixedCantileverBeam", types=str)
        self.options.declare("P", 50e3, types=(int, float), desc="Tip load (50 kN)")
        self.options.declare(
            "E", 200e9, types=(int, float), desc="Modulus of elast. (200 GPa)"
        )

    def _setup(self):
        self.options["ndim"] = 3
        self.listI = [
            0.0833,
            0.139,
            0.380,
            0.0796,
            0.133,
            0.363,
            0.0859,
            0.136,
            0.360,
            0.0922,
            0.138,
            0.369,
        ]

        self._set_design_space(
            DesignSpace(
                [
                    CategoricalVariable(values=[str(i + 1) for i in range(12)]),
                    FloatVariable(10.0, 20.0),
                    FloatVariable(1.0, 2.0),
                ]
            )
        )

    def _evaluate(self, x, kx=0):
        """
        Arguments
        ---------
        x : ndarray[ne, nx]
            Evaluation points.

        Returns
        -------
        ndarray[ne, 1]
            Functions values.
        """
        P = self.options["P"]
        E = self.options["E"]
        I = np.int64(x[:, 0]) - 1
        L = x[:, 1]
        S = x[:, 2]
        Ival = np.array([self.listI[i] for i in I])
        y = (P * L**3) / (3 * E * S**2 * Ival)
        return y



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>
        
This package is distributed under New BSD license.
"""


class OptionsDictionary(object):
    """
    Generalization of the dictionary that allows for declaring keys.

    Attributes
    ----------
    _dict : dict
        Dictionary of option values keyed by option names.
    _declared_entries : dict
        Dictionary of declared entries.
    """

    def __init__(self):
        self._dict = {}
        self._declared_entries = {}

    def clone(self):
        """
        Return a clone of this object.

        Returns
        -------
        OptionsDictionary
            Deep-copied clone.
        """
        clone = self.__class__()
        clone._dict = dict(self._dict)
        clone._declared_entries = dict(self._declared_entries)
        return clone

    def __getitem__(self, name):
        """
        Get an option that was previously declared and optionally set.

        Arguments
        ---------
        name : str
            The name of the option.

        Returns
        -------
        object
            Value of the option.
        """
        return self._dict[name]

    def __setitem__(self, name, value):
        """
        Set an option that was previously declared.

        The value argument must be valid, which means it must satisfy the following:
        1. If values and not types was given when declaring, value must be in values.
        2. If types and not values was given when declaring, type(value) must be in types.
        3. If values and types were given when declaring, either of the above must be true.

        Arguments
        ---------
        name : str
            The name of the option.
        value : object
            The value to set.
        """
        assert name in self._declared_entries, "Option %s has not been declared" % name
        self._assert_valid(name, value)
        self._dict[name] = value

    def __contains__(self, key):
        return key in self._dict

    def is_declared(self, key):
        return key in self._declared_entries

    def _assert_valid(self, name, value):
        values = self._declared_entries[name]["values"]
        types = self._declared_entries[name]["types"]

        if values is not None and types is not None:
            assert value in values or isinstance(
                value, types
            ), "Option %s: value and type of %s are both invalid - " % (
                name,
                value,
            ) + "value must be %s or type must be %s" % (
                values,
                types,
            )
        elif values is not None:
            assert value in values, "Option %s: value %s is invalid - must be %s" % (
                name,
                value,
                values,
            )
        elif types is not None:
            assert isinstance(
                value, types
            ), "Option %s: type of %s is invalid - must be %s" % (name, value, types)

    def update(self, dict_):
        """
        Loop over and set all the entries in the given dictionary into self.

        Arguments
        ---------
        dict_ : dict
            The given dictionary. All keys must have been declared.
        """
        for name in dict_:
            self[name] = dict_[name]

    def declare(self, name, default=None, values=None, types=None, desc=""):
        """
        Declare an option.

        The value of the option must satisfy the following:
        1. If values and not types was given when declaring, value must be in values.
        2. If types and not values was given when declaring, type(value) must be in types.
        3. If values and types were given when declaring, either of the above must be true.

        Arguments
        ---------
        name : str
            Name of the option.
        default : object
            Optional default value that must be valid under the above 3 conditions.
        values : list
            Optional list of acceptable option values.
        types : type or list of types
            Optional list of acceptable option types.
        desc : str
            Optional description of the option.
        """
        self._declared_entries[name] = {
            "values": values,
            "types": types,
            "default": default,
            "desc": desc,
        }

        if default is not None:
            self._assert_valid(name, default)

        self._dict[name] = default



from .ls import LS
from .qp import QP
from .krg import KRG
from .kpls import KPLS
from .gekpls import GEKPLS
from .kplsk import KPLSK
from .genn import GENN
from .mgp import MGP
from .sgp import SGP

from .krg_based import MixIntKernelType
from smt.utils.design_space import (
    DesignSpace,
    FloatVariable,
    IntegerVariable,
    OrdinalVariable,
    CategoricalVariable,
)
from smt.utils.kriging import MixHrcKernelType

try:
    from .idw import IDW
    from .rbf import RBF
    from .rmtc import RMTC
    from .rmtb import RMTB
except:
    pass



"""
G R A D I E N T - E N H A N C E D   N E U R A L   N E T W O R K S  (G E N N)

Author: Steven H. Berguin <steven.berguin@gtri.gatech.edu>

This package is distributed under New BSD license.
"""

from smt.surrogate_models.surrogate_model import SurrogateModel
from smt.utils.neural_net.model import Model

import numpy as np


# ------------------------------------ S U P P O R T   F U N C T I O N S -----------------------------------------------


def load_smt_data(model, xt, yt, dyt_dxt=None):
    """
    Utility function to load SMT data more easily

    :param model: SurrogateModel object for which to load training data
    :param xt: smt data points at which response is evaluated
    :param yt: response at xt
    :param dyt_dxt: gradient at xt
    """
    # Dimensionality
    if len(xt.shape) == 1:
        n_x = 1  # number of variables, x
        m = xt.size
    else:
        m, n_x = xt.shape

    if len(yt.shape) == 1:
        n_y = 1  # number of responses, y
    else:
        n_y = yt.shape[1]

    # Reshape arrays
    xt = xt.reshape((m, n_x))
    yt = yt.reshape((m, n_y))

    # Load values
    model.set_training_values(xt, yt)

    # Load partials
    if dyt_dxt is not None:
        dyt_dxt = dyt_dxt.reshape((m, n_x))
        for i in range(n_x):
            model.set_training_derivatives(xt, dyt_dxt[:, i].reshape((m, 1)), i)


def smt_to_genn(training_points):
    """
    Translate from SMT data structure to GENN data structure.

    Concretely, this neural net module works with numpy arrays in the form of (X, Y, J) as defined here-under. However,
    SMT uses a different format. Hence, we need a function that takes care of the translation.

    :param: training_points -- dict, training data in the format used by surrogate_model.py (see SMT API)

    Returns:
    :return X -- a numpy matrix of input features of shape (n_x, m) where n_x = no. of inputs, m = no. of train examples
    :return Y -- a numpy matrix of output labels of shape (n_y, m) where n_y = no. of outputs
    :return J -- a numpy array of size (n_y, n_x, m) representing the Jacobian of Y w.r.t. X:

        dY1/dX1 = J[0][0][:]
        dY1/dX2 = J[0][1][:]
        ...
        dY2/dX1 = J[1][0][:]
        dY2/dX2 = J[1][1][:]
        ...

        N.B. To retrieve the i^th example for dY2/dX1: J[1][0][i] for all i = 1,...,m

    """
    # Retrieve training data from SMT training_points
    xt, yt = training_points[None][
        0
    ]  # training_points[name][0] = [np.array(xt), np.array(yt)]

    # Deduce number of dimensions and training examples
    m, n_x = xt.shape
    _, n_y = yt.shape

    # Assign training data but transpose to match neural net implementation
    X = xt
    Y = yt

    # Loop to retrieve each partial derivative from SMT training_points
    J = np.zeros((m, n_x, n_y))
    for k in range(0, n_x):
        xt, dyt_dxt = training_points[None][k + 1]

        # assert that dimensions match
        assert xt.shape[0] == m
        assert xt.shape[1] == n_x
        assert dyt_dxt.shape[0] == m
        assert dyt_dxt.shape[1] == n_y

        # Assert that derivatives provided are for the same training points
        assert xt.all() == X.all()

        # Assign training derivative but transpose to match neural net implementation
        J[:, k, :] = dyt_dxt

    return X.T, Y.T, J.T


# ------------------------------------ C L A S S -----------------------------------------------------------------------


class GENN(SurrogateModel):
    name = "GENN"

    def _initialize(self):
        """API function: set default values for user options"""
        declare = self.options.declare
        declare("alpha", 0.5, types=(int, float), desc="optimizer learning rate")
        declare(
            "beta1", 0.9, types=(int, float), desc="Adam optimizer tuning parameter"
        )
        declare(
            "beta2", 0.99, types=(int, float), desc="Adam optimizer tuning parameter"
        )
        declare("lambd", 0.1, types=(int, float), desc="regularization coefficient")
        declare(
            "gamma", 1.0, types=(int, float), desc="gradient-enhancement coefficient"
        )
        declare("deep", 2, types=int, desc="number of hidden layers")
        declare("wide", 2, types=int, desc="number of nodes per hidden layer")
        declare(
            "mini_batch_size",
            64,
            types=int,
            desc="split data into batches of specified size",
        )
        declare(
            "num_epochs", 10, types=int, desc="number of random passes through the data"
        )
        declare(
            "num_iterations",
            100,
            types=int,
            desc="number of optimizer iterations per mini-batch",
        )
        declare(
            "seed",
            None,
            types=int,
            desc="random seed to ensure repeatability of results when desired",
        )
        declare("is_print", True, types=bool, desc="print progress (or not)")

        self.supports["derivatives"] = True
        self.supports["training_derivatives"] = True

        self.model = Model()

        self._is_trained = False

    def _train(self):
        """
        API function: train the neural net
        """
        # Convert training data to format expected by neural net module
        X, Y, J = smt_to_genn(self.training_points)

        # If there are no training derivatives, turn off gradient-enhancement
        if type(J) == np.ndarray and J.size == 0:
            self.options["gamma"] = 0.0

        # Get hyperparameters from SMT API
        alpha = self.options["alpha"]
        beta1 = self.options["beta1"]
        beta2 = self.options["beta2"]
        lambd = self.options["lambd"]
        gamma = self.options["gamma"]
        deep = self.options["deep"]
        wide = self.options["wide"]
        mini_batch_size = self.options["mini_batch_size"]
        num_iterations = self.options["num_iterations"]
        num_epochs = self.options["num_epochs"]
        seed = self.options["seed"]
        is_print = self.options["is_print"]

        # number of inputs and outputs
        n_x = X.shape[0]
        n_y = Y.shape[0]

        # Train neural net
        self.model = Model.initialize(n_x, n_y, deep, wide)
        self.model.train(
            X=X,
            Y=Y,
            J=J,
            num_iterations=num_iterations,
            mini_batch_size=mini_batch_size,
            num_epochs=num_epochs,
            alpha=alpha,
            beta1=beta1,
            beta2=beta2,
            lambd=lambd,
            gamma=gamma,
            seed=seed,
            silent=not is_print,
        )

        self._is_trained = True

    def _predict_values(self, x):
        """
        API method: predict values using appropriate methods from the neural_network.py module

        :param  x: np.ndarray[n, nx] -- Input values for the prediction points
        :return y: np.ndarray[n, ny] -- Output values at the prediction points
        """
        return self.model.evaluate(x.T).T

    def _predict_derivatives(self, x, kx):
        """
        API method: predict partials using appropriate methods from the neural_network.py module

        :param  x: np.ndarray[n, nx] -- Input values for the prediction points
        :param kx: int -- The 0-based index of the input variable with respect to which derivatives are desired
        :return: dy_dx: np.ndarray[n, ny] -- partial derivatives
        """
        return self.model.gradient(x.T)[:, kx, :].T

    def plot_training_history(self):
        if self._is_trained:
            self.model.plot_training_history()

    def goodness_of_fit(self, xv, yv, dyv_dxv):
        """
        Compute metrics to evaluate goodness of fit and show actual by predicted plot

        :param xv: np.ndarray[n, nx], x validation points
        :param yv: np.ndarray[n, 1], y validation response
        :param dyv_dxv: np.ndarray[n, ny], dydx validation derivatives
        """
        # Store current training points
        training_points = self.training_points

        # Replace training points with test (validation) points
        load_smt_data(self, xv, yv, dyv_dxv)

        # Convert from SMT format to a more convenient format for GENN
        X, Y, J = smt_to_genn(self.training_points)

        # Generate goodness of fit plots
        self.model.goodness_of_fit(X, Y)

        # Restore training points
        self.training_points = training_points


def run_example(is_gradient_enhancement=True):  # pragma: no cover
    """Test and demonstrate GENN using a 1D example"""
    import matplotlib.pyplot as plt

    # Test function
    f = lambda x: x * np.sin(x)
    df_dx = lambda x: np.sin(x) + x * np.cos(x)

    # Domain
    lb = -np.pi
    ub = np.pi

    # Training data
    m = 4
    xt = np.linspace(lb, ub, m)
    yt = f(xt)
    dyt_dxt = df_dx(xt)

    # Validation data
    xv = lb + np.random.rand(30, 1) * (ub - lb)
    yv = f(xv)
    dyv_dxv = df_dx(xv)

    # Initialize GENN object
    genn = GENN()
    genn.options["alpha"] = 0.1
    genn.options["beta1"] = 0.9
    genn.options["beta2"] = 0.99
    genn.options["lambd"] = 0.1
    genn.options["gamma"] = int(is_gradient_enhancement)
    genn.options["deep"] = 2
    genn.options["wide"] = 6
    genn.options["mini_batch_size"] = 64
    genn.options["num_epochs"] = 20
    genn.options["num_iterations"] = 100
    genn.options["is_print"] = True

    # Load data
    load_smt_data(genn, xt, yt, dyt_dxt)

    # Train
    genn.train()
    genn.plot_training_history()
    genn.goodness_of_fit(xv, yv, dyv_dxv)

    # Plot comparison
    if genn.options["gamma"] == 1.0:
        title = "with gradient enhancement"
    else:
        title = "without gradient enhancement"
    x = np.arange(lb, ub, 0.01)
    y = f(x)
    y_pred = genn.predict_values(x)
    fig, ax = plt.subplots()
    ax.plot(x, y_pred)
    ax.plot(x, y, "k--")
    ax.plot(xv, yv, "ro")
    ax.plot(xt, yt, "k+", mew=3, ms=10)
    ax.set(xlabel="x", ylabel="y", title=title)
    ax.legend(["Predicted", "True", "Test", "Train"])
    plt.show()


if __name__ == "__main__":  # pragma: no cover
    run_example(is_gradient_enhancement=True)



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>
        Dr. John T. Hwang         <hwangjt@umich.edu>

Water flow problem from:
Xiong, S., Qian, P. Z., & Wu, C. J. (2013). Sequential design and analysis of high-accuracy and low-accuracy computer codes. Technometrics, 55(1), 37-46.
"""
import numpy as np
from scipy.misc import derivative

from smt.problems.problem import Problem


class WaterFlowLFidelity(Problem):
    def _initialize(self):
        self.options.declare("name", "WaterFlowLFidelity", types=str)
        self.options.declare("use_FD", False, types=bool)
        self.options["ndim"] = 8

    def _setup(self):
        assert self.options["ndim"] == 8, "ndim must be 8"

        self.xlimits[:, 0] = [0.05, 100, 63070, 990, 63.1, 700, 1120, 9855]
        self.xlimits[:, 1] = [0.15, 50000, 115600, 1110, 116, 820, 1680, 12045]

    def _evaluate(self, x, kx):
        """
        Arguments
        ---------
        x : ndarray[ne, nx]
            Evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[ne, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        ne, nx = x.shape

        y = np.zeros((ne, 1), complex)

        def partial_derivative(function, var=0, point=[]):
            args = point[:]

            def wraps(x):
                args[var] = x
                return func(*args)

            return derivative(wraps, point[var], dx=1e-6)

        def func(x0, x1, x2, x3, x4, x5, x6, x7):
            return (
                5
                * x2
                * (x3 - x5)
                / (
                    np.log(x1 / x0)
                    * (1.5 + 2 * x6 * x2 / (np.log(x1 / x0) * x0**2 * x7) + x2 / x4)
                )
            )

        for i in range(ne):
            x0 = x[i, 0]
            x1 = x[i, 1]
            x2 = x[i, 2]
            x3 = x[i, 3]
            x4 = x[i, 4]
            x5 = x[i, 5]
            x6 = x[i, 6]
            x7 = x[i, 7]
            if kx is None:
                y[i, 0] = func(x0, x1, x2, x3, x4, x5, x6, x7)
            else:
                point = [x0, x1, x2, x3, x4, x5, x6, x7]
                if self.options["use_FD"]:
                    point = np.real(np.array(point))
                    y[i, 0] = partial_derivative(func, var=kx, point=point)
                else:
                    ch = 1e-20
                    point[kx] += complex(0, ch)
                    y[i, 0] = np.imag(func(*point)) / ch
                    point[kx] -= complex(0, ch)

        return y



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>
        Dr. John T. Hwang         <hwangjt@umich.edu>

This package is distributed under New BSD license.

Torsion vibration problem from:
Liu, H., Xu, S., & Wang, X. Sampling strategies and metamodeling techniques for engineering design: comparison and application. In ASME Turbo Expo 2016: Turbomachinery Technical Conference and Exposition. American Society of Mechanical Engineers. June, 2016.
Wang, L., Beeson, D., Wiggs, G., and Rayasam, M. A Comparison of Metamodeling Methods Using Practical Industry Requirements. In Proceedings of the 47th AIAA/ASME/ASCE/AHS/ASC structures, structural dynamics, and materials conference, Newport, RI, pp. AIAA 2006-1811.
"""
import numpy as np
from scipy.misc import derivative

from smt.problems.problem import Problem


class TorsionVibration(Problem):
    def _initialize(self):
        self.options.declare("name", "TorsionVibration", types=str)
        self.options.declare("use_FD", False, types=bool)
        self.options["ndim"] = 15

    def _setup(self):
        assert self.options["ndim"] == 15, "ndim must be 15"

        self.xlimits[:, 0] = [
            1.8,
            9,
            10530000,
            7.2,
            3510000,
            10.8,
            1.6425,
            10.8,
            5580000,
            2.025,
            2.7,
            0.252,
            12.6,
            3.6,
            0.09,
        ]
        self.xlimits[:, 1] = [
            2.2,
            11,
            12870000,
            8.8,
            4290000,
            13.2,
            2.0075,
            13.2,
            6820000,
            2.475,
            3.3,
            0.308,
            15.4,
            4.4,
            0.11,
        ]

    def _evaluate(self, x, kx):
        """
        Arguments
        ---------
        x : ndarray[ne, nx]
            Evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[ne, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        ne, nx = x.shape

        y = np.zeros((ne, 1), complex)

        def partial_derivative(function, var=0, point=[]):
            args = point[:]

            def wraps(x):
                args[var] = x
                return func(*args)

            return derivative(wraps, point[var], dx=1e-6)

        def func(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14):
            K1 = np.pi * x2 * x0 / (32 * x1)
            K2 = np.pi * x8 * x6 / (32 * x7)
            K3 = np.pi * x4 * x9 / (32 * x3)
            M1 = x11 * np.pi * x10 * x5 / (4 * 9.80665)
            M2 = x14 * np.pi * x13 * x12 / (4 * 9.80665)
            J1 = 0.5 * M1 * (x5 / 2) ** 2
            J2 = 0.5 * M2 * (x12 / 2) ** 2
            a = 1
            b = -((K1 + K2) / J1 + (K2 + K3) / J2)
            c = (K1 * K2 + K2 * K3 + K3 * K1) / (J1 * J2)
            return np.sqrt((-b - np.sqrt(b**2 - 4 * a * c)) / (2 * a)) / (2 * np.pi)

        for i in range(ne):
            x0 = x[i, 0]
            x1 = x[i, 1]
            x2 = x[i, 2]
            x3 = x[i, 3]
            x4 = x[i, 4]
            x5 = x[i, 5]
            x6 = x[i, 6]
            x7 = x[i, 7]
            x8 = x[i, 8]
            x9 = x[i, 9]
            x10 = x[i, 10]
            x11 = x[i, 11]
            x12 = x[i, 12]
            x13 = x[i, 13]
            x14 = x[i, 14]
            if kx is None:
                y[i, 0] = func(
                    x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14
                )
            else:
                point = [
                    x0,
                    x1,
                    x2,
                    x3,
                    x4,
                    x5,
                    x6,
                    x7,
                    x8,
                    x9,
                    x10,
                    x11,
                    x12,
                    x13,
                    x14,
                ]
                if self.options["use_FD"]:
                    point = np.real(np.array(point))
                    y[i, 0] = partial_derivative(func, var=kx, point=point)
                else:
                    ch = 1e-20
                    point[kx] += complex(0, ch)
                    y[i, 0] = np.imag(func(*point)) / ch
                    point[kx] -= complex(0, ch)

        return y



"""
Author: Remi Lafage <remi.lafage@onera.fr>

This package is distributed under New BSD license.

Branin function.
"""
import numpy as np

from smt.problems.problem import Problem


class Branin(Problem):
    def _initialize(self):
        self.options.declare("ndim", 2, values=[2], types=int)
        self.options.declare("name", "Branin", types=str)

    def _setup(self):
        assert self.options["ndim"] == 2, "ndim must be 2"

        self.xlimits[0, :] = [-5.0, 10]
        self.xlimits[1, :] = [0.0, 15]

    def _evaluate(self, x, kx):
        """
        Arguments
        ---------
        x : ndarray[ne, 2]
            Evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[ne, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        ne, nx = x.shape
        assert nx == 2, "x.shape[1] must be 2"

        y = np.zeros((ne, 1), complex)
        b = 5.1 / (4.0 * (np.pi) ** 2)
        c = 5.0 / np.pi
        t = 1.0 / (8.0 * np.pi)
        u = x[:, 1] - b * x[:, 0] ** 2 + c * x[:, 0] - 6
        if kx is None:
            r = 10.0 * (1.0 - t) * np.cos(x[:, 0]) + 10
            y[:, 0] = u**2 + r
        else:
            assert kx in [0, 1], "kx must be None, 0 or 1"
            if kx == 0:
                du_dx0 = -2 * b * x[:, 0] + c
                dr_dx0 = -10.0 * (1.0 - t) * np.sin(x[:, 0])
                y[:, 0] = 2 * du_dx0 * u + dr_dx0
            else:  # kx == 1
                y[:, 0] = 2 * u

        return y



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

Base class for sampling algorithms.
"""
from abc import ABCMeta, abstractmethod
import numpy as np

from smt.utils.options_dictionary import OptionsDictionary


class SamplingMethod(metaclass=ABCMeta):
    def __init__(self, **kwargs):
        """
        Constructor where values of options can be passed in.

        For the list of options, see the documentation for the problem being used.

        Parameters
        ----------
        **kwargs : named arguments
            Set of options that can be optionally set; each option must have been declared.

        Examples
        --------
        >>> import numpy as np
        >>> from smt.sampling_methods import Random
        >>> sampling = Random(xlimits=np.arange(2).reshape((1, 2)))
        """
        self.options = OptionsDictionary()
        self.options.declare(
            "xlimits",
            types=np.ndarray,
            desc="The interval of the domain in each dimension with shape nx x 2 (required)",
        )
        self._initialize(**kwargs)
        self.options.update(kwargs)

    def _initialize(self, **kwargs) -> None:
        """
        Implemented by sampling methods to declare options
        and/or use these optional values for initialization (optional)

        Parameters
        ----------
        **kwargs : named arguments passed by the user
            Set of options that can be optionally set

        Examples
        --------
        self.options.declare('option_name', default_value, types=(bool, int), desc='description')
        """
        pass

    def __call__(self, nt: int) -> np.ndarray:
        """
        Compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        return self._compute(nt)

    @abstractmethod
    def _compute(self, nt: int) -> np.ndarray:
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        raise Exception("This sampling method has not been implemented correctly")


class ScaledSamplingMethod(SamplingMethod):
    """This class describes an sample method which generates samples in the unit hypercube.

    The __call__ method does scale the generated samples accordingly to the defined xlimits.
    """

    def __call__(self, nt: int) -> np.ndarray:
        """
        Compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        return _scale_to_xlimits(self._compute(nt), self.options["xlimits"])

    @abstractmethod
    def _compute(self, nt: int) -> np.ndarray:
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the unit hypercube.
        """
        raise Exception("This sampling method has not been implemented correctly")


def _scale_to_xlimits(samples: np.ndarray, xlimits: np.ndarray) -> np.ndarray:
    """Scales the samples from the unit hypercube to the specified limits.

    Parameters
    ----------
    samples : np.ndarray
        The samples with coordinates in [0,1]
    xlimits : np.ndarray
        The xlimits

    Returns
    -------
    np.ndarray
        The scaled samples.
    """
    nx = xlimits.shape[0]
    for kx in range(nx):
        samples[:, kx] = xlimits[kx, 0] + samples[:, kx] * (
            xlimits[kx, 1] - xlimits[kx, 0]
        )
    return samples



from libcpp.vector cimport vector
from cpython cimport array
import numpy as np
cimport numpy as np


cdef extern from "rmtb.hpp":
  cdef cppclass RMTB:
    RMTB() except +
    void setup(int nx, double * lower, double * upper, int * order_list, int * ncp_list)
    void compute_ext_dist(int n, int nterm, double * x, double * dx)
    void compute_quadrature_points(int n, int * nelem_list, double * x)
    void compute_jac(int ix1, int ix2, int n, double * t, double * data, int * rows, int * cols)

cdef extern from "rmtc.hpp":
  cdef cppclass RMTC:
    RMTC() except +
    void setup(int nx, double * lower, double * upper, int * nelem_list, int * nterm_list)
    void compute_ext_dist(int n, int nterm, double * x, double * dx)
    void compute_quadrature_points(int n, int * nelem_list, double * x)
    void compute_coeff2nodal(double * mtx)
    void compute_uniq2elem(double * data, int * rows, int * cols)
    void compute_full_from_block(double * mtx, double * data, int * rows, int * cols)
    void compute_jac(int ix1, int ix2, int n, double * x, double * data, int * rows, int * cols)

cdef class PyRMTB:

    cdef RMTB *thisptr
    def __cinit__(self):
        self.thisptr = new RMTB()
    def __dealloc__(self):
        del self.thisptr
    def setup(self, int nx,
            np.ndarray[double] lower, np.ndarray[double] upper,
            np.ndarray[int] order_list, np.ndarray[int] ncp_list):
        self.thisptr.setup(nx, &lower[0], &upper[0], &order_list[0], &ncp_list[0])
    def compute_ext_dist(self, int n, int nterm, np.ndarray[double] x, np.ndarray[double] dx):
        self.thisptr.compute_ext_dist(n, nterm, &x[0], &dx[0])
    def compute_quadrature_points(self, int n, np.ndarray[int] nelem_list, np.ndarray[double] x):
        self.thisptr.compute_quadrature_points(n, &nelem_list[0], &x[0])
    def compute_jac(self, int ix1, int ix2, int n, np.ndarray[double] t,
            np.ndarray[double] data, np.ndarray[int] rows, np.ndarray[int] cols):
        self.thisptr.compute_jac(ix1, ix2, n, &t[0], &data[0], &rows[0], &cols[0])

cdef class PyRMTC:

    cdef RMTC *thisptr
    def __cinit__(self):
        self.thisptr = new RMTC()
    def __dealloc__(self):
        del self.thisptr
    def setup(self, int nx,
            np.ndarray[double] lower, np.ndarray[double] upper,
            np.ndarray[int] nelem_list, np.ndarray[int] nterm_list):
        self.thisptr.setup(nx, &lower[0], &upper[0], &nelem_list[0], &nterm_list[0])
    def compute_ext_dist(self, int n, int nterm, np.ndarray[double] x, np.ndarray[double] dx):
        self.thisptr.compute_ext_dist(n, nterm, &x[0], &dx[0])
    def compute_quadrature_points(self, int n, np.ndarray[int] nelem_list, np.ndarray[double] x):
        self.thisptr.compute_quadrature_points(n, &nelem_list[0], &x[0])
    def compute_coeff2nodal(self, np.ndarray[double] mtx):
        self.thisptr.compute_coeff2nodal(&mtx[0])
    def compute_uniq2elem(self,
            np.ndarray[double] data, np.ndarray[int] rows, np.ndarray[int] cols):
        self.thisptr.compute_uniq2elem(&data[0], &rows[0], &cols[0])
    def compute_full_from_block(self, np.ndarray[double] mtx,
            np.ndarray[double] data, np.ndarray[int] rows, np.ndarray[int] cols):
        self.thisptr.compute_full_from_block(&mtx[0], &data[0], &rows[0], &cols[0])
    def compute_jac(self, int ix1, int ix2, int n, np.ndarray[double] x,
            np.ndarray[double] data, np.ndarray[int] rows, np.ndarray[int] cols):
        self.thisptr.compute_jac(ix1, ix2, n, &x[0], &data[0], &rows[0], &cols[0])



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

Random sampling.
"""
import numpy as np

from smt.sampling_methods.sampling_method import ScaledSamplingMethod


class Random(ScaledSamplingMethod):
    def _compute(self, nt):
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the unit hypercube.
        """
        xlimits = self.options["xlimits"]
        nx = xlimits.shape[0]
        return np.random.rand(nt, nx)



"""
Author: P.Saves 
This package is distributed under New BSD license.

Multi-Layer Perceptron problem from:
C. Audet, E. Hall e-Hannan, and S. Le Digabel. A general mathematical framework for constrained mixed-variable blackbox optimization problems with meta and categorical variables. Operations Research Forum,499
4:137, 2023.
    """
import numpy as np

from smt.problems.problem import Problem
from smt.utils.design_space import (
    DesignSpace,
    OrdinalVariable,
    FloatVariable,
    CategoricalVariable,
    IntegerVariable,
)


class HierarchicalNeuralNetwork(Problem):
    def _initialize(self):
        self.options.declare("name", "HierarchicalNeuralNetwork", types=str)

    def _setup(self):
        design_space = DesignSpace(
            [
                OrdinalVariable(values=[1, 2, 3]),  # x0
                FloatVariable(-5, 2),
                FloatVariable(-5, 2),
                OrdinalVariable(values=[8, 16, 32, 64, 128, 256]),  # x3
                CategoricalVariable(values=["ReLU", "SELU", "ISRLU"]),  # x4
                IntegerVariable(0, 5),  # x5
                IntegerVariable(0, 5),  # x6
                IntegerVariable(0, 5),  # x7
            ]
        )

        # x6 is active when x0 >= 2
        design_space.declare_decreed_var(decreed_var=6, meta_var=0, meta_value=[2, 3])
        # x7 is active when x0 >= 3
        design_space.declare_decreed_var(decreed_var=7, meta_var=0, meta_value=3)

        self._set_design_space(design_space)

    def _evaluate(self, x, kx=0):
        """
        Arguments
        ---------
        x : ndarray[ne, nx]
            Evaluation points.

        Returns
        -------
        ndarray[ne, 1]
            Functions values.
        """
        ds = self.design_space

        def f_neu(x1, x2, x3, x4):
            if x4 == "ReLU":
                return 2 * x1 + x2 - 0.5 * x3
            elif x4 == "SELU":
                return -x1 + 2 * x2 - 0.5 * x3
            elif x4 == "ISRLU":
                return -x1 + x2 + 0.5 * x3
            else:
                raise ValueError(f"Unexpected x4: {x4}")

        def f1(x1, x2, x3, x4, x5):
            return f_neu(x1, x2, x3, x4) + x5**2

        def f2(x1, x2, x3, x4, x5, x6):
            return f_neu(x1, x2, x3, x4) + (x5**2) + 0.3 * x6

        def f3(x1, x2, x3, x4, x5, x6, x7):
            return f_neu(x1, x2, x3, x4) + (x5**2) + 0.3 * x6 - 0.1 * x7**3

        def f(X):
            y = []
            x0_decoded = ds.decode_values(X, i_dv=0)
            x3_decoded = ds.decode_values(X, i_dv=3)
            x4_decoded = ds.decode_values(X, i_dv=4)
            for i, x in enumerate(X):
                x0 = x0_decoded[i]
                x3 = x3_decoded[i]
                x4 = x4_decoded[i]
                if x0 == 1:
                    y.append(f1(x[1], x[2], x3, x4, x[5]))
                elif x0 == 2:
                    y.append(f2(x[1], x[2], x3, x4, x[5], x[6]))
                elif x0 == 3:
                    y.append(f3(x[1], x[2], x3, x4, x[5], x[6], x[7]))
                else:
                    raise ValueError(f"Unexpected x0 value: {x0}")
            return np.array(y)

        return f(x)



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

Full-factorial sampling.
"""
import numpy as np

from smt.sampling_methods.sampling_method import ScaledSamplingMethod


class FullFactorial(ScaledSamplingMethod):
    def _initialize(self, **kwargs):
        self.options.declare(
            "weights",
            values=None,
            types=(list, np.ndarray),
            desc="relative sampling weights for each nx dimensions",
        )
        self.options.declare(
            "clip",
            default=False,
            types=bool,
            desc="round number of samples to the sampling number product of each nx dimensions (> asked nt)",
        )

    def _compute(self, nt):
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the unit hypercube.
        """
        xlimits = self.options["xlimits"]
        nx = xlimits.shape[0]

        if self.options["weights"] is None:
            weights = np.ones(nx) / nx
        else:
            weights = np.atleast_1d(self.options["weights"])
            weights /= np.sum(weights)

        num_list = np.ones(nx, int)
        while np.prod(num_list) < nt:
            ind = np.argmax(weights - num_list / np.sum(num_list))
            num_list[ind] += 1

        lins_list = [np.linspace(0.0, 1.0, num_list[kx]) for kx in range(nx)]
        x_list = np.meshgrid(*lins_list, indexing="ij")

        if self.options["clip"]:
            nt = np.prod(num_list)

        x = np.zeros((nt, nx))
        for kx in range(nx):
            x[:, kx] = x_list[kx].reshape(np.prod(num_list))[:nt]

        return x



"""
G R A D I E N T - E N H A N C E D   N E U R A L   N E T W O R K S  (G E N N)

Author: Steven H. Berguin <steven.berguin@gtri.gatech.edu>

This package is distributed under New BSD license.
"""

import numpy as np


def rsquare(Y_pred, Y_true):
    """
    Compute R-square for a single response.

    NOTE: If you have more than one response, then you'll either have to modify this method to handle many responses at
          once or wrap a for loop around it (i.e. treat one response at a time).

    Arguments:
    Y_pred -- predictions,  numpy array of shape (K, m) where n_y = no. of outputs, m = no. of examples
    Y_true -- true values,  numpy array of shape (K, m) where n_y = no. of outputs, m = no. of examples

    Return:
    R2 -- the R-square value,  numpy array of shape (K, 1)
    """
    epsilon = 1e-8  # small number to avoid division by zero
    Y_bar = np.mean(Y_true)
    SSE = np.sum(np.square(Y_pred - Y_true), axis=1)
    SSTO = np.sum(np.square(Y_true - Y_bar) + epsilon, axis=1)
    R2 = 1 - SSE / SSTO
    return R2



"""
G R A D I E N T - E N H A N C E D   N E U R A L   N E T W O R K S  (G E N N)

Author: Steven H. Berguin <steven.berguin@gtri.gatech.edu>

This package is distributed under New BSD license.
"""

import numpy as np

EPS = np.finfo(float).eps  # small number to avoid division by zero


def initialize_back_prop(AL, Y, AL_prime, Y_prime):
    """
    Initialize backward propagation

    Arguments:
    :param AL -- output of the forward propagation L_model_forward()... i.e. neural net predictions   (if regression)
                                                                 i.e. neural net probabilities (if classification)
          >> a numpy array of shape (n_y, m) where n_y = no. outputs, m = no. examples

    :param Y -- true "label" (classification) or "value" (regression)
         >> a numpy array of shape (n_y, m) where n_y = no. outputs, m = no. examples

    :param AL_prime -- the derivative of the last layer's activation output(s) w.r.t. the inputs x: AL' = d(AL)/dX
                >> a numpy array of size (n_y, n_x, m) where n_y = no. outputs, n_x = no. inputs, m = no. examples

    :param Y_prime -- the true derivative of the output(s) w.r.t. the inputs x: Y' = d(Y)/dX
               >> a numpy array of shape (n_y, n_x, m) where n_y = no. outputs, n_x = no. inputs, m = no. examples

    Returns:
    :return dAL -- gradient of the loss function w.r.t. last layer activations: d(L)/dAL
           >> a numpy array of shape (n_y, m)

    :return dAL_prime -- gradient of the loss function w.r.t. last layer activations derivatives: d(L)/dAL' where AL' = d(AL)/dX
                 >> a numpy array of shape (n_y, n_x, m)
    """
    n_y, _ = AL.shape  # number layers, number examples
    Y = Y.reshape(AL.shape)
    dAL = AL - Y  # derivative of loss function w.r.t. to activations: dAL = d(L)/dAL
    dAL_prime = (
        AL_prime - Y_prime
    )  # derivative of loss function w.r.t. to partials: dAL_prime = d(L)/d(AL_prime)

    return dAL, dAL_prime


def linear_activation_backward(dA, dA_prime, cache, J_cache, lambd, gamma):
    """
    Implement backward propagation for one LINEAR->ACTIVATION layer for the regression least squares estimation

    Arguments:
    :param dA -- post-activation gradient w.r.t. A for current layer l, dA = d(L)/dA where L is the loss function
            >> a numpy array of shape (n_1, m) where n_l = no. nodes in current layer, m = no. of examples

    :param dA_prime -- post-activation gradient w.r.t. A' for current layer l, dA' = d(L)/dA' where L is the loss function
                                                                                        and A' = d(AL) / dX
            >> a numpy array of shape (n_l, n_x, m) where n_l = no. nodes in current layer
                                                          n_x = no. of inputs (X1, X2, ...)
                                                          m = no. of examples

    :param cache -- tuple of values stored in linear_activation_forward()
              >> a tuple containing (A_prev, Z, W, b, activation)
                      where
                            A_prev -- activations from previous layer
                                      >> a numpy array of shape (n_prev, m) where n_prev is the no. nodes in layer L-1
                            Z -- input to activation functions for current layer
                                      >> a numpy array of shape (n, m) where n is the no. nodes in layer L
                            W -- weight parameters for current layer
                                 >> a numpy array of shape (n, n_prev)
                            b -- bias parameters for current layer
                                 >> a numpy array of shape (n, 1)
                            activation -- activation function to use

    :param J_cache -- list of caches containing every cache of L_grads_forward() where J stands for Jacobian
              >> a list containing [..., (j, Z_prime_j, A_prime_j, G_prime, G_prime_prime), ...]
                                          ------------------ input j --------------------
                      where
                            j -- input variable associated with current cache
                                      >> an integer representing the associated input variables (X1, X2, ..., Xj, ...)
                            Z_prime_j -- derivative of Z w.r.t. X_j: Z'_j = d(Z_j)/dX_j
                                      >> a numpy array of shape (n_l, m) where n_l is the no. nodes in layer l
                            A_prime_j -- derivative of the activation w.r.t. X_j: A'_j = d(A_j)/dX_j
                                      >> a numpy array of shape (n_l, m) where n_l is the no. nodes in layer l

    :param lambd: float, regularization parameter
    :param gamma: float, gradient-enhancement parameter

    :return dA_prev -- Gradient of the cost with respect to the activation (of the previous layer l-1), same shape as A_prev
    :return dW -- Gradient of the cost with respect to W (current layer l), same shape as W
    :return db -- Gradient of the cost with respect to b (current layer l), same shape as b
    """
    # Extract information from current layer cache (avoids recomputing what was previously computed)
    A_prev, Z, W, b, activation = cache

    # Some dimensions that will be useful
    m = A_prev.shape[1]  # number of examples
    n = len(J_cache)  # number of inputs

    # 1st derivative of activation function A = G(Z)
    G_prime = activation.first_derivative(Z)

    # Compute the contribution due to the 0th order terms (where regularization only affects dW)
    dW = (
        1.0 / m * np.dot(G_prime * dA, A_prev.T) + lambd / m * W
    )  # dW = d(J)/dW where J is the cost function
    db = 1.0 / m * np.sum(G_prime * dA, axis=1, keepdims=True)  # db = d(J)/db
    dA_prev = np.dot(
        W.T, G_prime * dA
    )  # dA_prev = d(L)/dA_prev where A_prev = previous layer activation

    # Initialize dA_prime_prev = d(J)/dA_prime_prev
    dA_prime_prev = np.zeros((W.shape[1], n, m))

    # Gradient enhancement
    if gamma != 0:
        # 2nd derivative of activation function A = G(Z)
        G_prime_prime = activation.second_derivative(Z)

        # Loop over partials, d()/dX_j
        for j_cache in J_cache:
            # Extract information from current layer cache associated with derivative of A w.r.t. j^th input
            j, Z_prime_j, A_prime_j_prev = j_cache

            # Extract partials of A w.r.t. to j^th input, i.e. A_prime_j = d(A)/dX_j
            dA_prime_j = dA_prime[:, j, :].reshape(Z_prime_j.shape)

            # Compute contribution to cost function gradient, db = d(J)/db, dW = d(J)/dW, d(L)/dA, d(L)/dA'
            dW += (
                gamma
                / m
                * (
                    np.dot(dA_prime_j * G_prime_prime * Z_prime_j, A_prev.T)
                    + np.dot(dA_prime_j * G_prime, A_prime_j_prev.T)
                )
            )
            db += (
                gamma
                / m
                * np.sum(dA_prime_j * G_prime_prime * Z_prime_j, axis=1, keepdims=True)
            )
            dA_prev += gamma * np.dot(W.T, dA_prime_j * G_prime_prime * Z_prime_j)
            dA_prime_prev[:, j, :] = gamma * np.dot(W.T, dA_prime_j * G_prime)

    return dA_prev, dW, db, dA_prime_prev


def L_model_backward(AL, Y, AL_prime, Y_prime, caches, J_caches, lambd, gamma):
    """
    Implement backward propagation

    Arguments:
    :param AL -- output of the forward propagation L_model_forward()... i.e. neural net predictions   (if regression)
                                                                 i.e. neural net probabilities (if classification)
          >> a numpy array of shape (n_y, m) where n_y = no. outputs, m = no. examples

    :param Y -- true "label" (classification) or "value" (regression)
         >> a numpy array of shape (n_y, m) where n_y = no. outputs, m = no. examples

    :param AL_prime -- the derivative of the last layer's activation output(s) w.r.t. the inputs x: AL' = d(AL)/dX
                >> a numpy array of size (n_y, n_x, m) where n_y = no. outputs, n_x = no. inputs, m = no. examples

    :param Y_prime -- the true derivative of the output(s) w.r.t. the inputs x: Y' = d(Y)/dX
               >> a numpy array of shape (n_y, n_x, m) where n_y = no. outputs, n_x = no. inputs, m = no. examples

    :param caches -- list of caches containing every cache of L_model_forward()
              >> a tuple containing {(A_prev, Z, W, b, activation), ..., (A_prev, Z, W, b, activation)}
                                      -------- layer 1 -----------        -------- layer L ----------
                      where
                            A_prev -- activations from previous layer
                                      >> a numpy array of shape (n_prev, m) where n_prev is the no. nodes in layer L-1
                            Z -- input to activation functions for current layer
                                      >> a numpy array of shape (n, m) where n is the no. nodes in layer L
                            W -- weight parameters for current layer
                                 >> a numpy array of shape (n, n_prev)
                            b -- bias parameters for current layer
                                 >> a numpy array of shape (n, 1)

    :param J_caches -- a list of lists containing every cache of L_grads_forward() for each layer (where J stands for Jacobian)
              >> a tuple [ [[...], ..., [...]], ..., [..., (j, Z_prime_j, A_prime_j, G_prime, G_prime_prime), ...], ...]
                            --- layer 1 ------        ------------------ layer l, partial j ---------------------
                      where
                            j -- input variable number (i.e. X1, X2, ...) associated with cache
                                      >> an integer representing the associated input variables (X1, X2, ..., Xj, ...)
                            Z_prime_j -- derivative of Z w.r.t. X_j: Z'_j = d(Z_j)/dX_j
                                      >> a numpy array of shape (n_l, m) where n_l is the no. nodes in layer l
                            A_prime_j -- derivative of the activation w.r.t. X_j: A'_j = d(A_j)/dX_j
                                      >> a numpy array of shape (n_l, m) where n_l is the no. nodes in layer l

    :param lambd: float, regularization parameter
    :param gamma: float, gradient-enhancement parameter


    :return grads -- A dictionary with the gradients of the cost function w.r.t. to parameters:
                grads["A" + str(l)] = ...
                grads["W" + str(l)] = ...
                grads["b" + str(l)] = ...
    """
    # Initialize grads
    grads = {}

    # Some quantities needed
    L = len(caches)  # the number of layers
    _, m = AL.shape
    Y = Y.reshape(AL.shape)  # after this line, Y is the same shape as AL

    # Initializing the back propagation
    dA, dA_prime = initialize_back_prop(AL, Y, AL_prime, Y_prime)

    # Loop over each layer
    for l in reversed(range(L)):
        # Get cache
        cache = caches[l]
        J_cache = J_caches[l]

        # Backprop step
        dA, dW, db, dA_prime = linear_activation_backward(
            dA, dA_prime, cache, J_cache, lambd, gamma
        )

        # Store result
        grads["W" + str(l + 1)] = dW
        grads["b" + str(l + 1)] = db

    return grads



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>
       Dr. John T. Hwang         <hwangjt@umich.edu>

This package is distributed under New BSD license.

Welded beam problem from:
Liu, H., Xu, S., & Wang, X. Sampling strategies and metamodeling techniques for engineering design: comparison and application. In ASME Turbo Expo 2016: Turbomachinery Technical Conference and Exposition. American Society of Mechanical Engineers. June, 2016.
Deb, K. An Efficient Constraint Handling Method for Genetic Algorithms. Computer methods in applied mechanics and engineering, 186(2), pp. 311-338. 2000.
"""
import numpy as np
from scipy.misc import derivative
from smt.problems.problem import Problem


class WeldedBeam(Problem):
    def _initialize(self):
        self.options.declare("name", "WeldedBeam", types=str)
        self.options.declare("use_FD", False, types=bool)
        self.options["ndim"] = 3

    def _setup(self):
        assert self.options["ndim"] == 3, "ndim must be 3"  # t, h, l
        self.xlimits[:, 0] = [5, 0.125, 5]
        self.xlimits[:, 1] = [10, 1, 10]

    def _evaluate(self, x, kx):
        """
        Arguments
        ---------
        x : ndarray[ne, nx]
            Evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[ne, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        ne, nx = x.shape
        y = np.zeros((ne, 1), complex)

        def partial_derivative(function, var=0, point=[]):
            args = point[:]

            def wraps(x):
                args[var] = x
                return func(*args)

            return derivative(wraps, point[var], dx=1e-6)

        def func(x0, x1, x2):
            tau1 = 6000 / (np.sqrt(2) * x1 * x2)
            tau2 = (
                6000
                * (14 + 0.5 * x2)
                * np.sqrt(0.25 * (x2**2 + (x1 + x0) ** 2))
                / (2 * (0.707 * x1 * x2 * (x2 / 12.0 + 0.25 * (x1 + x0) ** 2)))
            )
            return np.sqrt(
                tau1**2
                + tau2**2
                + x2 * tau1 * tau2 / np.sqrt(0.25 * (x2**2 + (x1 + x0) ** 2))
            )

        for i in range(ne):
            x0 = x[i, 0]
            x1 = x[i, 1]
            x2 = x[i, 2]
            if kx is None:
                y[i, 0] = func(x0, x1, x2)
            else:
                point = [x0, x1, x2]
                if self.options["use_FD"]:
                    point = np.real(np.array(point))
                    y[i, 0] = partial_derivative(func, var=kx, point=point)
                else:
                    ch = 1e-20
                    point[kx] += complex(0, ch)
                    y[i, 0] = np.imag(func(*point)) / ch
                    point[kx] -= complex(0, ch)

        return y



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>, R. Lafage

This package is distributed under New BSD license.

Variable-fidelity modeling: two types of bridge functions are available; i.e.,
additive and multiplicative
"""

import numpy as np

from smt.utils.options_dictionary import OptionsDictionary
from types import FunctionType
from smt.applications.application import SurrogateBasedApplication


class VFM(SurrogateBasedApplication):
    def _initialize(self):
        super(VFM, self)._initialize()

        declare = self.options.declare

        declare(
            "name_model_LF",
            types=object,
            values=(
                "KRG",
                "LS",
                "QP",
                "KPLS",
                "KPLSK",
                "GEKPLS",
                "RBF",
                "RMTC",
                "RMTB",
                "IDW",
            ),
            desc="Name of the low-fidelity model",
        )
        declare("options_LF", {}, types=dict, desc="Options for the low-fidelity model")
        declare(
            "name_model_bridge",
            types=object,
            values=(
                "KRG",
                "LS",
                "QP",
                "KPLS",
                "KPLSK",
                "GEKPLS",
                "RBF",
                "RMTC",
                "RMTB",
                "IDW",
            ),
            desc="Name of the bridge model",
        )
        declare("options_bridge", {}, types=dict, desc="Options for the bridge model")
        declare(
            "type_bridge",
            "Additive",
            types=str,
            values=("Additive", "Multiplicative"),
            desc="Bridge function type",
        )
        declare("X_LF", None, types=np.ndarray, desc="Low-fidelity inputs")
        declare("y_LF", None, types=np.ndarray, desc="Low-fidelity output")
        declare("X_HF", None, types=np.ndarray, desc="High-fidelity inputs")
        declare("y_HF", None, types=np.ndarray, desc="High-fidelity output")
        declare("dy_LF", None, types=np.ndarray, desc="Low-fidelity derivatives")
        declare("dy_HF", None, types=np.ndarray, desc="High-fidelity derivatives")

        self.nx = None
        self.ny = None
        self.sm_HF = None
        self._trained = False

    def predict_values(self, x):
        """
        Predict the output values at a set of points x.

        Parameters
        ----------
        x: np.ndarray[n, nx] or np.ndarray[n]
           Input values for the prediction result analysis.

        kx : int
           The 0-based index of the input variable with respect to which derivatives are desired.

        return
        ------
        y: np.ndarray
            Output values at the prediction points.

        """
        if not self._trained:
            self._apply()
        y = self.sm_HF["predict_values"](x)
        return y

    def predict_derivatives(self, x, kx):
        """
        Predict the dy_dx derivatives at a set of points.

        Parameters
        ----------
        x: np.ndarray[n, nx] or np.ndarray[n]
           Input values for the prediction result analysis.

        kx : int
           The 0-based index of the input variable with respect to which derivatives are desired.

        return
        ------
        y: np.ndarray
            Derivatives at the prediction points.

        """
        if not self._trained:
            self._apply()
        if kx is None:
            y = np.zeros(x.shape)
            for i in range(x.shape[1]):
                y[:, i] = self.sm_HF["predict_derivatives"][i](x).reshape((x.shape[0]))
        else:
            y = self.sm_HF["predict_derivatives"][kx](x).reshape((x.shape[0], self.ny))

        return y

    def _apply(self):
        """
        Algorithm of the VFM method
        """

        # For seek of readability
        if (
            self.options["X_LF"] is not None
            and self.options["y_LF"] is not None
            and self.options["X_HF"] is not None
            and self.options["y_HF"] is not None
        ):
            X_LF = self.options["X_LF"]
            y_LF = self.options["y_LF"]
            X_HF = self.options["X_HF"]
            y_HF = self.options["y_HF"]
        else:
            raise ValueError("Check X_LF, y_LF, X_HF, and y_FH")

        self.nx = X_LF.shape[1]
        self.ny = y_LF.shape[1]

        if self.options["dy_LF"] is not None:
            dy_LF = self.options["dy_LF"]
        if self.options["dy_HF"] is not None:
            dy_HF = self.options["dy_HF"]

        # Check parameters
        self._check_param()

        # Train the low fidelity model
        self.LF_deriv = self.options["options_LF"]["deriv"]
        del self.options["options_LF"]["deriv"]
        sm_LF = self.options["name_model_LF"](**self.options["options_LF"])
        sm_LF.options["print_global"] = False
        sm_LF.set_training_values(X_LF, y_LF)
        if self.LF_deriv:
            for i in range(sm_LF.nx):
                sm_LF.set_training_derivatives(X_LF, dy_LF[:, i], i)

        sm_LF.train()

        # compute the bridge data
        if self.options["type_bridge"] == "Multiplicative":
            y_bridge = y_HF / sm_LF.predict_values(X_HF)
            self.B_deriv = self.options["options_bridge"]["deriv"]
            del self.options["options_bridge"]["deriv"]
            if self.B_deriv:
                dy_bridge = np.zeros(dy_HF.shape)
                for i in range(sm_LF.nx):
                    dy_bridge[:, i] = (
                        (
                            (
                                dy_HF[:, i].reshape((y_HF.shape[0], 1))
                                * sm_LF.predict_values(X_HF)
                            )
                            - (y_HF * sm_LF.predict_derivatives(X_HF, kx=i))
                        )
                        / sm_LF.predict_values(X_HF) ** 2
                    ).reshape(y_HF.shape[0])
        elif self.options["type_bridge"] == "Additive":
            y_bridge = y_HF - sm_LF.predict_values(X_HF)
            self.B_deriv = self.options["options_bridge"]["deriv"]
            del self.options["options_bridge"]["deriv"]
            if self.B_deriv:
                dy_bridge = np.zeros(dy_HF.shape)
                for i in range(sm_LF.nx):
                    dy_bridge[:, i] = (
                        dy_HF[:, i].reshape((y_HF.shape[0], 1))
                        - sm_LF.predict_derivatives(X_HF, kx=i)
                    ).reshape((y_HF.shape[0]))
        else:
            raise ValueError("Only Additive and Multiplicative bridges are available")

        # Construct of the bridge function
        sm_bridge = self.options["name_model_bridge"](**self.options["options_bridge"])
        sm_bridge.set_training_values(X_HF, y_bridge)
        if self.B_deriv and self.options["dy_HF"] is not None:
            for i in range(sm_bridge.nx):
                sm_bridge.set_training_derivatives(X_HF, dy_bridge[:, i], i)
        sm_bridge.train()

        # Construct the final model
        sm_HF = {}
        if self.options["type_bridge"] == "Multiplicative":
            sm_HF["predict_values"] = lambda x: sm_bridge.predict_values(
                x
            ) * sm_LF.predict_values(x)
            if sm_bridge.supports["derivatives"] and sm_LF.supports["derivatives"]:
                sm_HF["predict_derivatives"] = []
                for i in range(sm_LF.nx):
                    sm_HF["predict_derivatives"].append(
                        lambda x, i=i: sm_bridge.predict_derivatives(x, i)
                        * sm_LF.predict_values(x)
                        + sm_bridge.predict_values(x) * sm_LF.predict_derivatives(x, i)
                    )

        else:
            sm_HF["predict_values"] = lambda x: sm_bridge.predict_values(
                x
            ) + sm_LF.predict_values(x)
            if sm_bridge.supports["derivatives"] and sm_LF.supports["derivatives"]:
                sm_HF["predict_derivatives"] = []
                for i in range(sm_LF.nx):
                    sm_HF["predict_derivatives"].append(
                        lambda x, i=i: sm_bridge.predict_derivatives(x, i)
                        + sm_LF.predict_derivatives(x, i)
                    )

        self._trained = True
        self.sm_HF = sm_HF

    def _check_param(self):
        """
        This function check some parameters of the model.
        """
        # Check surrogates
        if not callable(self.options["name_model_LF"]):
            if self.options["name_model_LF"] in self._surrogate_type:
                self.options["name_model_LF"] = self._surrogate_type[
                    self.options["name_model_LF"]
                ]
            else:
                raise ValueError(
                    "The LF surrogate should be one of %s, "
                    "%s was given."
                    % (self._surrogate_type.keys(), self.options["name_model_LF"])
                )

        if not callable(self.options["name_model_bridge"]):
            if self.options["name_model_bridge"] in self._surrogate_type:
                self.options["name_model_bridge"] = self._surrogate_type[
                    self.options["name_model_bridge"]
                ]
            else:
                raise ValueError(
                    "The bridge surrogate should be one of %s, "
                    "%s was given."
                    % (self._surrogate_type.keys(), self.options["name_model_bridge"])
                )

        # Initialize the parameter deriv
        if "deriv" not in self.options["options_LF"].keys():
            self.options["options_LF"]["deriv"] = False
        if "deriv" not in self.options["options_bridge"].keys():
            self.options["options_bridge"]["deriv"] = False
